using System;
using System.Threading;
using System.Windows.Forms;
using SlimDX;
using SlimDX.DirectInput;

namespace PolledDevices
{
	public abstract class PolledDevice
    {
        #region Fields 

        protected SlimDX.DirectInput.DirectInput _dinput;
        protected SlimDX.DirectInput.Device _device;
        
        protected Control _owner;
        
        private bool _acquired;
		private long _pollTime;
		private System.Threading.Timer _pollTimer;

        #endregion

        #region Events

        public event EventHandler DeviceAcquiredEvent;
		public event EventHandler DeviceReleasedEvent;
		public event EventHandler DeviceLostEvent;

        #endregion

        /// <summary>
        /// 
        /// </summary>
        /// <param name="owner"></param>
		public PolledDevice(Control owner)
		{
            // set this property to allow us to run as we did under VS .NET 2003. it allows
            // us to call across threads; that is, to call into the owner's thread even
            // though we don't own the 'owner', and it runs in a main UI thread.
            Control.CheckForIllegalCrossThreadCalls = false;
            _dinput = null;
			_owner = owner;
			_pollTime = 10; // ms
		}

        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
		public override string ToString()
		{
			return (_device != null) ? _device.Information.ProductName : base.ToString();
		}

        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
		protected abstract bool AcquireDevice();
		
        /// <summary>
        /// 
        /// </summary>
        protected virtual void ReleaseDevice() {}

        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        protected abstract bool CompareState();

        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
		public bool Acquire()
		{
			if (_acquired)
				return true;

            try
            {
                _dinput = new DirectInput(); // set up direct input here
                _acquired = AcquireDevice(); // call into subclass
            }
            catch (SlimDX.DirectInput.DirectInputException ex)
            {
                MessageBox.Show(this._owner, ex.Message, "Device Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                _acquired = false;
            }

			if (_acquired)
			{
				if (DeviceAcquiredEvent != null)
					DeviceAcquiredEvent(this, EventArgs.Empty);

				// start the lightweight polling timer
				_pollTimer = new System.Threading.Timer(new TimerCallback(OnPollDevice), this, _pollTime, Timeout.Infinite);
			}
			return _acquired;
		}

        /// <summary>
        /// 
        /// </summary>
		public void Release()
		{
			if (!_acquired)
				return;
            _acquired = false;

			if (_pollTimer != null)
			{
				_pollTimer.Change(Timeout.Infinite, Timeout.Infinite);
				_pollTimer.Dispose();
				_pollTimer = null;
			}

			if (_device != null)
			{
                try
                {
                    _device.Unacquire();
                    _device.Dispose();
                }
                catch (SlimDX.DirectInput.DirectInputException)
                {
                    // do nothing
                }
                finally
                {
                    _device = null;
                }
			}

            // give any inherited class a chance to release additional resources
			ReleaseDevice(); // call into subclass

            // release the resources required for direct input
            if (_dinput != null)
            {
                _dinput.Dispose();
                _dinput = null;
            }

            // fire the released event
			if (DeviceReleasedEvent != null)
				DeviceReleasedEvent(this, EventArgs.Empty);
		}

        /// <summary>
        /// 
        /// </summary>
		public bool IsAcquired
		{
			get { return _acquired; }
		}

        /// <summary>
        /// Opens the built-in control panel GUI for SlimDX.
        /// </summary>
		public virtual void Calibrate()
		{
            if (_device != null)
                _device.RunControlPanel(_owner);
		}

        /// <summary>
        /// 
        /// </summary>
        public long PollTime
        {
            get
            {
                return _pollTime;
            }
            set
            {
                if (value < 1L)
                    value = 1L;
                _pollTime = value;

                if (_acquired && _pollTimer != null) // restart timer
                    _pollTimer.Change(_pollTime, Timeout.Infinite);
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="state"></param>
		private static void OnPollDevice(object state)
		{
			PolledDevice This = (PolledDevice) state;
            
            This._pollTimer.Change(Timeout.Infinite, Timeout.Infinite);
			
            if (This == null || This._device == null || !This._acquired || This._owner.IsDisposed)
				return;

			try
			{
				// readies the next batch of data
                This._device.Poll();
			}
            catch (SlimDX.DirectInput.DirectInputException)
			{
                try
                {
                    Thread.Sleep(1000); // wait one second before retrying
                    This._device.Acquire(); // try to re-acquire
                    This._acquired = true; // just to be sure
                    if (This.DeviceAcquiredEvent != null)
                        This.DeviceAcquiredEvent(This, EventArgs.Empty);
                }
                catch (SlimDX.DirectInput.DirectInputException)
                {
                    This._acquired = false;
                    if (This.DeviceLostEvent != null)
                        This.DeviceLostEvent(This, EventArgs.Empty);
                    return; // could not acquire
                }
			}

			if (This.CompareState()) // call into subclasses to fire events!
			{
				if (This._pollTimer != null)
				{
					// re-start the timer and poll for more data
					This._pollTimer.Change(This._pollTime, Timeout.Infinite);
				}
			}
			else
			{
				// failed to acquire the state for comparison
				This._acquired = false;
				if (This.DeviceLostEvent != null)
					This.DeviceLostEvent(This, EventArgs.Empty);
			}
		}
	}
}
